@@ -7,13 +7,12 @@ Note: All ID variables used in the methods of this class are global IDs, unless
 """
 import hashlib
 from copy import deepcopy
+import sys
 from typing import Dict
-
 import numpy as np
 from mpi4py import MPI
 
 from .adaptivity import AdaptivityCalculator
-from ..micro_simulation import create_simulation_class
 
 
 class GlobalAdaptivityCalculator(AdaptivityCalculator):
@@ -23,8 +22,10 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
         global_number_of_sims: int,
         global_ids: list,
         participant,
+        base_logger,
         rank: int,
-        comm_world,
+        comm,
+        micro_problem_cls,
     ) -> None:
         """
         Class constructor.
@@ -39,39 +40,32 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
             List of global IDs of simulations living on this rank.
         participant : object of class Participant
             Object of the class Participant using which the preCICE API is called.
+        base_logger : object of class Logger
+            Logger object to log messages.
         rank : int
             MPI rank.
-        comm_world : MPI.COMM_WORLD
-            Base global communicator of MPI.
+        comm : MPI.Comm
+            Communicator for MPI.
+        micro_problem_cls : callable
+            Class of micro problem.
         """
-        super().__init__(configurator, rank, global_number_of_sims)
+        super().__init__(
+            configurator, global_number_of_sims, micro_problem_cls, base_logger, rank
+        )
         self._global_number_of_sims = global_number_of_sims
         self._global_ids = global_ids
-        self._comm_world = comm_world
-
-        local_number_of_sims = len(global_ids)
-
-        # Create a map of micro simulation global IDs and the ranks on which they are
-        micro_sims_on_this_rank = np.zeros(local_number_of_sims, dtype=np.intc)
-        for i in range(local_number_of_sims):
-            micro_sims_on_this_rank[i] = self._rank
+        self._comm = comm
 
         rank_of_sim = self._get_ranks_of_sims()
 
         self._is_sim_on_this_rank = [False] * global_number_of_sims  # DECLARATION
-        for i in range(global_number_of_sims):
-            if rank_of_sim[i] == self._rank:
-                self._is_sim_on_this_rank[i] = True
+        for gid in range(global_number_of_sims):
+            if rank_of_sim[gid] == self._rank:
+                self._is_sim_on_this_rank[gid] = True
 
         self._precice_participant = participant
 
-        if (
-            self._adaptivity_output_type == "all"
-            or self._adaptivity_output_type == "local"
-        ):
-            self._metrics_logger.log_info("n,n active,n inactive,assoc ranks")
-
-        self._comm_node = comm_world.Split_type(MPI.COMM_TYPE_SHARED)
+        self._comm_node = comm.Split_type(MPI.COMM_TYPE_SHARED)
 
         self._MPI_local_rank = self._comm_node.Get_rank()
 
@@ -128,10 +122,6 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
         data_for_adaptivity : dict
             Dictionary with keys as names of data to be used in the similarity calculation, and values as the respective data for the micro simulations
         """
-        self._precice_participant.start_profiling_section(
-            "global_adaptivity.compute_adaptivity"
-        )
-
         for name in data_for_adaptivity.keys():
             if name not in self._adaptivity_data_names:
                 raise ValueError(
@@ -143,8 +133,21 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
         # Gather adaptivity data from all ranks
         global_data_for_adaptivity = dict()
         for name in data_for_adaptivity.keys():
-            data_as_list = self._comm_world.allgather(data_for_adaptivity[name])
-            global_data_for_adaptivity[name] = np.concatenate((data_as_list[:]), axis=0)
+            data_as_list = self._comm.allgather(data_for_adaptivity[name])
+            global_ids_as_list = self._comm.allgather(self._global_ids)
+            global_data_for_adaptivity[name] = [0] * self._global_number_of_sims
+            for i, gids_list in enumerate(global_ids_as_list):
+                count = 0
+                for gid in gids_list:
+                    global_data_for_adaptivity[name][gid] = data_as_list[i][count]
+                    count += 1
+
+            # global_data_for_adaptivity[name] = np.concatenate(global_data_for_adaptivity[name], axis=0)
+            global_data_for_adaptivity[name] = np.array(
+                global_data_for_adaptivity[name]
+            )
+
+            # global_data_for_adaptivity[name] = np.concatenate((data_as_list[:]), axis=0)
 
         if (
             self._MPI_local_rank == 0
@@ -161,33 +164,69 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
 
         self._associate_inactive_to_active()
 
-        self._precice_participant.stop_last_profiling_section()
-
-    def get_active_sim_ids(self) -> np.ndarray:
+    def get_active_sim_local_ids(self) -> np.ndarray:
         """
-        Get the ids of active simulations.
+        Get the local ids of active simulations on this rank.
 
         Returns
         -------
         numpy array
             1D array of active simulation ids
         """
-        return np.where(
-            self._is_sim_active[self._global_ids[0] : self._global_ids[-1] + 1]
-        )[0]
+        active_sim_ids = []
+        for gid in self._global_ids:
+            if self._is_sim_active[gid]:
+                active_sim_ids.append(self._global_ids.index(gid))
 
-    def get_inactive_sim_ids(self) -> np.ndarray:
+        return np.array(active_sim_ids)
+
+    def get_inactive_sim_local_ids(self) -> np.ndarray:
         """
-        Get the ids of inactive simulations.
+        Get the local ids of inactive simulations on this rank.
 
         Returns
         -------
         numpy array
             1D array of inactive simulation ids
         """
-        return np.where(
-            self._is_sim_active[self._global_ids[0] : self._global_ids[-1] + 1] == False
-        )[0]
+        inactive_sim_ids = []
+        for gid in self._global_ids:
+            if not self._is_sim_active[gid]:
+                inactive_sim_ids.append(self._global_ids.index(gid))
+
+        return np.array(inactive_sim_ids)
+
+    def get_active_sim_global_ids(self) -> np.ndarray:
+        """
+        Get the global ids of active simulations on this rank.
+
+        Returns
+        -------
+        numpy array
+            1D array of active simulation ids
+        """
+        active_sim_ids = []
+        for gid in self._global_ids:
+            if self._is_sim_active[gid]:
+                active_sim_ids.append(gid)
+
+        return np.array(active_sim_ids)
+
+    def get_inactive_sim_global_ids(self) -> np.ndarray:
+        """
+        Get the global ids of inactive simulations on this rank.
+
+        Returns
+        -------
+        numpy array
+            1D array of inactive simulation ids
+        """
+        inactive_sim_ids = []
+        for gid in self._global_ids:
+            if not self._is_sim_active[gid]:
+                inactive_sim_ids.append(gid)
+
+        return np.array(inactive_sim_ids)
 
     def get_full_field_micro_output(self, micro_output: list) -> list:
         """
@@ -204,7 +243,7 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
             List of dicts having individual output of each simulation. Active and inactive simulation outputs are entered.
         """
         self._precice_participant.start_profiling_section(
-            "global_adaptivity.get_full_field_micro_output"
+            "micro_manager.global_adaptivity.get_full_field_micro_output"
         )
 
         micro_sims_output = deepcopy(micro_output)
@@ -226,6 +265,8 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
 
         Global metrics:
         - Time window at which the metrics are logged
+        - Global number of active simulations
+        - Global number of inactive simulations
         - Average number of active simulations
         - Average number of inactive simulations
         - Maximum number of active simulations
@@ -238,8 +279,8 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
         """
         active_sims_on_this_rank = 0
         inactive_sims_on_this_rank = 0
-        for global_id in self._global_ids:
-            if self._is_sim_active[global_id]:
+        for gid in self._global_ids:
+            if self._is_sim_active[gid]:
                 active_sims_on_this_rank += 1
             else:
                 inactive_sims_on_this_rank += 1
@@ -251,16 +292,14 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
             ranks_of_sims = self._get_ranks_of_sims()
 
             assoc_ranks = []  # Ranks to which inactive sims on this rank are associated
-            for global_id in self._global_ids:
-                if not self._is_sim_active[global_id]:
-                    assoc_rank = int(
-                        ranks_of_sims[self._sim_is_associated_to[global_id]]
-                    )
+            for gid in self._global_ids:
+                if not self._is_sim_active[gid]:
+                    assoc_rank = int(ranks_of_sims[self._sim_is_associated_to[gid]])
                     if not assoc_rank in assoc_ranks:
                         assoc_ranks.append(assoc_rank)
 
             self._metrics_logger.log_info(
-                "{},{},{},{}".format(
+                "{}|{}|{}|{}".format(
                     n,
                     active_sims_on_this_rank,
                     inactive_sims_on_this_rank,
@@ -272,26 +311,66 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
             self._adaptivity_output_type == "all"
             or self._adaptivity_output_type == "global"
         ):
-            active_sims_rankwise = self._comm_world.gather(
-                active_sims_on_this_rank, root=0
-            )
-            inactive_sims_rankwise = self._comm_world.gather(
+            active_sims_rankwise = self._comm.gather(active_sims_on_this_rank, root=0)
+            inactive_sims_rankwise = self._comm.gather(
                 inactive_sims_on_this_rank, root=0
             )
 
             if self._rank == 0:
-                size = self._comm_world.Get_size()
+                size = self._comm.Get_size()
 
                 self._global_metrics_logger.log_info(
-                    "{},{},{},{},{}".format(
+                    "{}|{}|{}|{}|{}|{}|{}|{}|{}".format(
                         n,
+                        sum(active_sims_rankwise),
+                        sum(inactive_sims_rankwise),
                         sum(active_sims_rankwise) / size,
                         sum(inactive_sims_rankwise) / size,
                         max(active_sims_rankwise),
+                        active_sims_rankwise.index(max(active_sims_rankwise)),
                         max(inactive_sims_rankwise),
+                        inactive_sims_rankwise.index(max(inactive_sims_rankwise)),
                     )
                 )
 
+    def _update_active_sims(self) -> None:
+        """
+        Update set of active micro simulations.
+        Pairs of active simulations (A, B) are compared and if found to be similar, B is deactivated.
+        """
+        if self._max_similarity_dist == 0.0:
+            self._base_logger.log_warning(
+                "All similarity distances are zero, which means all the data for adaptivity is the same. Coarsening tolerance will be manually set to minimum float number."
+            )
+            self._coarse_tol = sys.float_info.min
+        else:
+            self._coarse_tol = (
+                self._coarse_const * self._refine_const * self._max_similarity_dist
+            )
+
+        active_gids_this_rank = self.get_active_sim_global_ids()
+        # Gather global ids of active sims from all ranks
+        active_gids_all_ranks = self._comm.allgather(active_gids_this_rank.tolist())
+
+        active_gids_to_iterate = []
+        # Iterate over global ids of active sims in a round-robin fashion across ranks
+        while any(len(gid_list) > 0 for gid_list in active_gids_all_ranks):
+            for gid_list in active_gids_all_ranks:
+                if gid_list:  # if the list of global ids is not empty
+                    # Pick the first global id on every rank and add it to the list which is later iterated over
+                    active_gids_to_iterate.append(gid_list[0])
+                    # Remove the picked global id from the rank list
+                    gid_list.pop(0)
+
+        # Update the set of active micro sims
+        active_gids_to_check = active_gids_to_iterate.copy()
+        for gid in active_gids_to_iterate:
+            if self._check_for_deactivation(gid, active_gids_to_check):
+                self._is_sim_active[gid] = False
+                self._just_deactivated.append(gid)
+                # Remove deactivated gid from further checks
+                active_gids_to_check.remove(gid)
+
     def _communicate_micro_output(
         self,
         micro_output: list,
@@ -305,40 +384,36 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
         micro_output : list
             List of dicts having individual output of each simulation. Only the active simulation outputs are entered.
         """
-
-        inactive_local_ids = self.get_inactive_sim_ids()
-
-        local_sim_is_associated_to = self._sim_is_associated_to[
-            self._global_ids[0] : self._global_ids[-1] + 1
-        ]
-
         # Keys are global IDs of active simulations associated to inactive
         # simulations on this rank. Values are global IDs of the inactive
         # simulations.
         active_to_inactive_map: Dict[int, list] = dict()
 
-        for i in inactive_local_ids:
-            assoc_active_id = local_sim_is_associated_to[i]
+        inactive_gids = self.get_inactive_sim_global_ids()
+
+        for gid in inactive_gids:
+            assoc_active_gid = self._sim_is_associated_to[gid]
             # Gather global IDs of associated active simulations not on this rank
-            if not self._is_sim_on_this_rank[assoc_active_id]:
-                if assoc_active_id in active_to_inactive_map:
-                    active_to_inactive_map[assoc_active_id].append(i)
+            if not self._is_sim_on_this_rank[assoc_active_gid]:
+                if assoc_active_gid in active_to_inactive_map:
+                    active_to_inactive_map[assoc_active_gid].append(gid)
                 else:
-                    active_to_inactive_map[assoc_active_id] = [i]
+                    active_to_inactive_map[assoc_active_gid] = [gid]
             else:  # If associated active simulation is on this rank, copy the output directly
-                micro_output[i] = deepcopy(
-                    micro_output[self._global_ids.index(assoc_active_id)]
-                )
+                lid = self._global_ids.index(gid)
+                assoc_active_lid = self._global_ids.index(assoc_active_gid)
+                micro_output[lid] = deepcopy(micro_output[assoc_active_lid])
 
-        assoc_active_ids = list(active_to_inactive_map.keys())
+        assoc_active_gids = list(active_to_inactive_map.keys())
 
-        recv_reqs = self._p2p_comm(assoc_active_ids, micro_output)
+        recv_reqs = self._p2p_comm(assoc_active_gids, micro_output)
 
         # Add received output of active sims to inactive sims on this rank
         for count, req in enumerate(recv_reqs):
             output = req.wait()
-            for local_id in active_to_inactive_map[assoc_active_ids[count]]:
-                micro_output[local_id] = deepcopy(output)
+            for gid in active_to_inactive_map[assoc_active_gids[count]]:
+                lid = self._global_ids.index(gid)
+                micro_output[lid] = deepcopy(output)
 
     def _update_inactive_sims(self, micro_sims: list) -> None:
         """
@@ -356,58 +431,58 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
 
         _sim_is_associated_to_updated = np.copy(self._sim_is_associated_to)
 
+        # -------------------- Global computation on every rank ----------------------------
         # Check inactive simulations for activation and collect IDs of those to be activated
-        to_be_activated_ids = []  # Global IDs to be activated
-        for i in range(self._is_sim_active.size):
-            if not self._is_sim_active[i]:  # if id is inactive
-                if self._check_for_activation(i, self._is_sim_active):
-                    self._is_sim_active[i] = True
-                    _sim_is_associated_to_updated[
-                        i
-                    ] = -2  # Active sim cannot have an associated sim
-                    if self._is_sim_on_this_rank[i] and i not in self._just_deactivated:
-                        to_be_activated_ids.append(i)
+        active_gids_all_ranks = np.where(self._is_sim_active)[0]
+        inactive_gids_all_ranks = np.where(self._is_sim_active == False)[0]
+        to_be_activated_gids = []  # Global IDs to be activated
+
+        for gid in inactive_gids_all_ranks:
+            if self._check_for_activation(gid, active_gids_all_ranks):
+                self._is_sim_active[gid] = True
+                # Active sim cannot have an associated sim
+                _sim_is_associated_to_updated[gid] = -2
+                if gid not in self._just_deactivated:
+                    # Add the newly activated gid to active_gids_all_ranks for further checks
+                    active_gids_all_ranks = np.append(active_gids_all_ranks, gid)
+                    # Collect the global IDs to be activated on this rank
+                    if self._is_sim_on_this_rank[gid]:
+                        to_be_activated_gids.append(gid)
+        # ----------------------------------------------------------------------------------
 
         self._just_deactivated.clear()  # Clear the list of sims deactivated in this step
 
-        local_sim_is_associated_to = self._sim_is_associated_to[
-            self._global_ids[0] : self._global_ids[-1] + 1
-        ]
-
         # Keys are global IDs of active sims not on this rank, values are lists of local and
         # global IDs of inactive sims associated to the active sims which are on this rank
         to_be_activated_map: Dict[int, list] = dict()
 
-        for i in to_be_activated_ids:
-            # Only handle activation of simulations on this rank -- LOCAL SCOPE HERE ON
-            if self._is_sim_on_this_rank[i]:
-                to_be_activated_local_id = self._global_ids.index(i)
-                micro_sims[to_be_activated_local_id] = create_simulation_class(
-                    self._micro_problem
-                )(i)
-                assoc_active_id = local_sim_is_associated_to[to_be_activated_local_id]
+        # Only handle activation of simulations on this rank
+        for gid in to_be_activated_gids:
+            to_be_activated_lid = self._global_ids.index(gid)
+            micro_sims[to_be_activated_lid] = self._micro_problem_cls(gid)
+            assoc_active_gid = self._sim_is_associated_to[gid]
 
-                if self._is_sim_on_this_rank[
-                    assoc_active_id
-                ]:  # Associated active simulation is on the same rank
-                    assoc_active_local_id = self._global_ids.index(assoc_active_id)
-                    micro_sims[to_be_activated_local_id].set_state(
-                        micro_sims[assoc_active_local_id].get_state()
-                    )
-                else:  # Associated active simulation is not on this rank
-                    if assoc_active_id in to_be_activated_map:
-                        to_be_activated_map[assoc_active_id].append(
-                            to_be_activated_local_id
-                        )
-                    else:
-                        to_be_activated_map[assoc_active_id] = [
-                            to_be_activated_local_id
-                        ]
+            if self._is_sim_on_this_rank[
+                assoc_active_gid
+            ]:  # Associated active simulation is on the same rank
+                assoc_active_lid = self._global_ids.index(assoc_active_gid)
+                micro_sims[to_be_activated_lid].set_state(
+                    micro_sims[assoc_active_lid].get_state()
+                )
+            else:  # Associated active simulation is not on this rank
+                if assoc_active_gid in to_be_activated_map:
+                    to_be_activated_map[assoc_active_gid].append(to_be_activated_lid)
+                else:
+                    to_be_activated_map[assoc_active_gid] = [to_be_activated_lid]
+
+        self._precice_participant.start_profiling_section(
+            "micro_manager.global_adaptivity.update_inactive_sims.communication"
+        )
 
         sim_states_and_global_ids = []
-        for local_id, sim in enumerate(micro_sims):
+        for lid, sim in enumerate(micro_sims):
             if sim == 0:
-                sim_states_and_global_ids.append((None, self._global_ids[local_id]))
+                sim_states_and_global_ids.append((None, self._global_ids[lid]))
             else:
                 sim_states_and_global_ids.append((sim.get_state(), sim.get_global_id()))
 
@@ -417,20 +492,20 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
 
         # Use received micro sims to activate the required simulations
         for req in recv_reqs:
-            state, global_id = req.wait()
-            local_ids = to_be_activated_map[global_id]
-            for local_id in local_ids:
+            state, gid = req.wait()
+            local_ids = to_be_activated_map[gid]
+            for lid in local_ids:
                 # Create the micro simulation object and set its state
-                micro_sims[local_id] = create_simulation_class(self._micro_problem)(
-                    self._global_ids[local_id]
-                )
-                micro_sims[local_id].set_state(state)
+                micro_sims[lid] = self._micro_problem_cls(self._global_ids[lid])
+                micro_sims[lid].set_state(state)
 
         # Delete the micro simulation object if it is inactive
-        for i in self._global_ids:
-            if not self._is_sim_active[i]:
-                local_id = self._global_ids.index(i)
-                micro_sims[local_id] = 0
+        for gid in self._global_ids:
+            if not self._is_sim_active[gid]:
+                lid = self._global_ids.index(gid)
+                micro_sims[lid] = 0
+
+        self._precice_participant.stop_last_profiling_section()
 
         self._sim_is_associated_to = np.copy(_sim_is_associated_to_updated)
 
@@ -497,7 +572,7 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
             send_map_local[i] = self._rank
 
         # Gather information about which sims to send where, from the sending perspective
-        send_map_list = self._comm_world.allgather(send_map_local)
+        send_map_list = self._comm.allgather(send_map_local)
 
         for d in send_map_list:
             for i, rank in d.items():
@@ -509,21 +584,21 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
 
         # Asynchronous send operations
         send_reqs = []
-        for global_id, send_ranks in send_map.items():
-            local_id = self._global_ids.index(global_id)
+        for gid, send_ranks in send_map.items():
+            lid = self._global_ids.index(gid)
             for send_rank in send_ranks:
-                tag = self._create_tag(global_id, self._rank, send_rank)
-                req = self._comm_world.isend(data[local_id], dest=send_rank, tag=tag)
+                tag = self._create_tag(gid, self._rank, send_rank)
+                req = self._comm.isend(data[lid], dest=send_rank, tag=tag)
                 send_reqs.append(req)
 
         # Asynchronous receive operations
         recv_reqs = []
-        for global_id, recv_rank in recv_map.items():
-            tag = self._create_tag(global_id, recv_rank, self._rank)
+        for gid, recv_rank in recv_map.items():
+            tag = self._create_tag(gid, recv_rank, self._rank)
             bufsize = (
                 1 << 30
             )  # allocate and use a temporary 1 MiB buffer size https://github.com/mpi4py/mpi4py/issues/389
-            req = self._comm_world.irecv(bufsize, source=recv_rank, tag=tag)
+            req = self._comm.irecv(bufsize, source=recv_rank, tag=tag)
             recv_reqs.append(req)
 
         # Wait for all non-blocking communication to complete
@@ -534,16 +609,17 @@ class GlobalAdaptivityCalculator(AdaptivityCalculator):
     def _get_ranks_of_sims(self) -> np.ndarray:
         """
         Get the ranks of all simulations.
+
         Returns
         -------
         ranks_of_sims : np.ndarray
             Array of ranks on which simulations exist.
         """
-        local_gids_to_rank = dict()
+        gids_to_rank = dict()
         for gid in self._global_ids:
-            local_gids_to_rank[gid] = self._rank
+            gids_to_rank[gid] = self._rank
 
-        ranks_maps_as_list = self._comm_world.allgather(local_gids_to_rank)
+        ranks_maps_as_list = self._comm.allgather(gids_to_rank)
 
         ranks_of_sims = np.zeros(self._global_number_of_sims, dtype=np.intc)
         for ranks_map in ranks_maps_as_list:
